#!/usr/bin/env python

# Copyright (c) 2009 J. Lewis Muir <jlmuir@anl.gov>
#
# Permission to use, copy, modify, and distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
# SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
# OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
# CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

#
# For usage, see printUsage function or run with -h option.
#

import getopt
import asynchat
import asyncore
import math
import os
import re
import signal
import socket
import sys
import time

DEFAULT_PORT = 50435

def main(args):
  (port, isDebugOutputEnabled) = parseCommandLineOptions(args)
  ConnectionDispatcher(port, isDebugOutputEnabled)
  asyncore.loop()

def getProgramName(args=None):
  if args == None: args = sys.argv
  if len(args) == 0 or args[0] == "-c": return "PROGRAM_NAME"
  return os.path.basename(args[0])

def initializeSignalHandlers():
  signalToNameMap = {
    signal.SIGHUP:"SIGHUP",
    signal.SIGINT:"SIGINT",
    signal.SIGPIPE:"SIGPIPE",
    signal.SIGALRM:"SIGALRM",
    signal.SIGTERM:"SIGTERM",
    signal.SIGUSR1:"SIGUSR1",
    signal.SIGUSR2:"SIGUSR2"
  }

  def signalHandler(signum, frame):
    if signum in [signal.SIGINT, signal.SIGTERM]:
      print
      print "%s caught. Exiting." % signalToNameMap[signum]
      sys.exit(0)
    else:
      print >> sys.stderr
      print >> sys.stderr, "%s caught. Exiting." % signalToNameMap[signum]
      sys.exit(1)

  for each in signalToNameMap.keys():
    signal.signal(each, signalHandler)

def parseCommandLineOptions(args):
  (options, extra) = getopt.getopt(args[1:], "dp:h", ["debug", "port=",
    "help"])

  if len(extra) > 0:
    print >> sys.stderr, "Error: unexpected command line option \"%s\"" % \
      extra[0]
    printUsage(sys.stderr)
    sys.exit(1)

  port = DEFAULT_PORT
  isDebugOutputEnabled = False

  for eachOptName, eachOptValue in options:
    if eachOptName in ("-d", "--debug"):
      isDebugOutputEnabled = True
    elif eachOptName in ("-p", "--port"):
      port = int(eachOptValue)
    elif eachOptName in ("-h", "--help"):
      printUsage(sys.stdout)
      sys.exit(0)

  return (port, isDebugOutputEnabled)

def printUsage(outStream):
  print >> outStream, """\
Usage: %s [OPTION]...
Options:
  -d,--debug        Print debug messages to stderr
  -p,--port=NUMBER  Listen on the specified port NUMBER for incoming
                    connections (default: %d)
  -h,--help         Print usage message and exit
""" % (getProgramName(), DEFAULT_PORT)


class ConnectionDispatcher(asyncore.dispatcher):
  def __init__(self, port, debugOutputEnabled=False):
    asyncore.dispatcher.__init__ (self)
    self.port = port
    self.debugOutputEnabled = debugOutputEnabled
    self.devices = []
    for i in xrange(6): self.devices.append(Tpg256A())
    self.create_socket(socket.AF_INET, socket.SOCK_STREAM)
    self.set_reuse_addr()
    self.bind(("", port))
    self.listen(5)

  def handle_accept(self):
    ConnectionSession(self.accept(), self.devices, self.debugOutputEnabled)


class ConnectionSession(asynchat.async_chat):
  # Simulation-only interface
  SetPrRe = re.compile(r"PR(?P<se>[1-6])=(?P<stat>\d)," + \
    r"(?P<press>\d\.\d{3}[eE][+-]\d+)$")

  # Device interface
  GetPrRe = re.compile(r"PR(?P<se>[1-6])$")

  def __init__(self, (conn, addr), devices, debugOutputEnabled=False):
    asynchat.async_chat.__init__(self, conn)
    self.devices = devices
    self.debugOutputEnabled = debugOutputEnabled
    self.set_terminator(None)
    self.commandTerminatorChar = "\r"       # must be string of length 1
    self.requestTransmissionChar = "\005" # must be string of length 1
    self.outputTerminator = "\r\n"
    self.commandAck = "\006"
    self.commandNak = "\025"
    self.mostRecentCommand = ""
    self.buffer = ""

  # Assumes self.commandTerminatorChar and self.requestTransmissionChar have
  # length == 1.  If this cannot be assumed, this implementation would need to
  # be made more complex to handle partial terminator sequences or partial
  # request transmission sequences in a read.
  def handle_read(self):
    try:
      data = self.recv(self.ac_in_buffer_size)
    except socket.error, e:
      self.handle_error()
      return

    commandTerminatorLen = len(self.commandTerminatorChar)
    requestTransmissionLen = len(self.requestTransmissionChar)

    while True:
      inpTermCharIdx = data.find(self.commandTerminatorChar)
      rqsTranCharIdx = data.find(self.requestTransmissionChar)
      if inpTermCharIdx == -1 and rqsTranCharIdx == -1:
        self.collect_incoming_data(data)
        break
      elif inpTermCharIdx == -1 or \
          (rqsTranCharIdx != -1 and rqsTranCharIdx < inpTermCharIdx):
        self.collect_incoming_data(data[:rqsTranCharIdx])
        self.found_transmission_request()
        data = data[rqsTranCharIdx+requestTransmissionLen:]
      else:
        self.collect_incoming_data(data[:inpTermCharIdx])
        self.found_command_terminator()
        data = data[inpTermCharIdx+commandTerminatorLen:]

  def collect_incoming_data(self, data):
    self.buffer = self.buffer + data

  def found_command_terminator(self):
    data = self.buffer
    self.buffer = ""
    if self.debugOutputEnabled:
      sys.stderr.write("< \"%s\"\n" % \
        (data + self.commandTerminatorChar).encode("string_escape"))
    self.handleClientRequest(data)
    self.mostRecentCommand = data

  def found_transmission_request(self):
    self.buffer = ""
    if self.debugOutputEnabled:
      sys.stderr.write("< \"%s\"\n" % \
        self.requestTransmissionChar.encode("string_escape"))
    self.handleTransmissionRequest(self.mostRecentCommand)

  def handleClientRequest(self, request):
    if self.GetPrRe.match(request):
      self.sendClientResponse(self.commandAck)
    elif self.SetPrRe.match(request):
      match = self.SetPrRe.match(request)
      sensor = int(match.group("se"))
      self.devices[sensor-1].status = int(match.group("stat"))
      self.devices[sensor-1].pressure = float(match.group("press"))
      self.sendClientResponse(self.commandAck)
    else:
      self.sendClientResponse(self.commandNak)

  def handleTransmissionRequest(self, request):
    if self.GetPrRe.match(request):
      match = self.GetPrRe.match(request)
      sensor = int(match.group("se"))
      # This will produce a response like "1.234E+00".  The device would
      # actually return a response like "1.234E+0" (only one exponent digit),
      # but as long as the driver doesn't assume this, this simulation will
      # be close enough and will work with the driver correctly.
      self.sendClientResponse("%d,%.3E" % \
        (self.devices[sensor-1].status,
         self.devices[sensor-1].pressure))

  def sendClientResponse(self, response):
    data = response + self.outputTerminator
    self.push(data)
    if self.debugOutputEnabled:
      sys.stderr.write("> \"%s\"\n" % data.encode("string_escape"))


class Tpg256A:
  def __init__(self):
    self.status = 0
    self.pressure = 1.0e-6


if __name__ == '__main__':
  try:
    initializeSignalHandlers()
    main(sys.argv)
  except Exception, e:
    if isinstance(e, SystemExit):
      raise e
    else:
      print >> sys.stderr
      print >> sys.stderr, "Error: %s" % e
      sys.exit(1)
